{% if router in ['fastapi', 'starlette'] %}
    {% include '_starlette_endpoints.py.jinja' %}
{% elif router == 'blacksheep' %}
    {% include '_blacksheep_endpoints.py.jinja' %}
{% elif router == 'xpresso' %}
    {% include '_xpresso_endpoints.py.jinja' %}
{% endif %}
